# encoding: utf8
#
# (C) Copyright Arskom Ltd. <info@arskom.com.tr>
#               Uğurcan Ergün <ugurcanergn@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

import logging
logger = logging.getLogger(__name__)

import sys

from twisted.internet import reactor
from twisted.web.server import Site
from twisted.web.wsgi import WSGIResource

from spyne.application import Application
from spyne.protocol.xml import XmlDocument
from spyne.protocol.html import HtmlTable
from spyne.protocol.http import HttpRpc
from spyne.server.wsgi import WsgiApplication

from spynepi.const import DB_CONNECTION_STRING
from spynepi.const import HOST
from spynepi.const import PORT
from spynepi.db import init_database
from spynepi.entity.html import IndexService
from spynepi.entity.html import HtmlService
from spynepi.entity.project import RdfService
from spynepi.entity.root import RootService

from werkzeug.exceptions import HTTPException
from werkzeug.routing import Map
from werkzeug.routing import Rule

from sqlalchemy.orm.exc import NoResultFound
from spyne.error import ResourceNotFoundError


def TWsgiApplication(url_map):
    def _application(environ, start_response, wsgi_url=None):
        urls = url_map.bind_to_environ(environ)
        try:
            endpoint, args = urls.match()
        except HTTPException, e:
            return e(environ, start_response)

        return endpoint(environ, start_response, wsgi_url)

    return _application


class MyApplication(Application):
    def call_wrapper(self, ctx):
        """This is the application-wide exception transforming function."""

        try:
            return Application.call_wrapper(self, ctx)

        except NoResultFound, e:
            logger.exception(e)
            ctx.out_string = ["Resource not found"]
            raise ResourceNotFoundError() # Return HTTP 404


def main(connection_string=DB_CONNECTION_STRING):
    # configure logging
    logging.basicConfig(level=logging.DEBUG)
    logging.getLogger('spyne.protocol.xml').setLevel(logging.DEBUG)
    #logging.getLogger('sqlalchemy.engine.base.Engine').setLevel(logging.DEBUG)

    index_app = MyApplication([RootService, IndexService], "http://usefulinc.com/ns/doap#",
                            in_protocol=HttpRpc(), out_protocol=HtmlTable())

    rdf_app = MyApplication([RdfService], "http://usefulinc.com/ns/doap#",
                            in_protocol=HttpRpc(), out_protocol=XmlDocument())

    html_app = MyApplication([HtmlService],"http://usefulinc.com/ns/doap#",
                            in_protocol=HttpRpc(), out_protocol=HttpRpc())

    db_handle = init_database(connection_string)

    class UserDefinedContext(object):
        def __init__(self):
            self.session = db_handle.Session()

        def close(self):
            self.session.close()

    # this is called after validation
    def _on_method_call(ctx):
        ctx.udc = UserDefinedContext()

    # this is called once all data is sent to the client.
    def _on_method_return_object(ctx):
        ctx.udc.session.commit()

    def _on_wsgi_close(ctx):
        if ctx.udc is not None:
            ctx.udc.close()

    for app in index_app, rdf_app, html_app:
        app.event_manager.add_listener('method_call', _on_method_call)
        app.event_manager.add_listener('method_return_object', _on_method_return_object)

    wsgi_index = WsgiApplication(index_app)
    wsgi_rdf = WsgiApplication(rdf_app)
    wsgi_html = WsgiApplication(html_app)

    for a in wsgi_index,wsgi_rdf,wsgi_html:
        a.event_manager.add_listener('wsgi_close', _on_wsgi_close)

    url_map = Map([Rule("/", endpoint=wsgi_index),
        Rule("/<project_name>", endpoint=wsgi_html),
        Rule("/<project_name>/", endpoint=wsgi_html),
        Rule("/<project_name>/doap.rdf",endpoint=wsgi_rdf),
        Rule("/<project_name>/<version>", endpoint=wsgi_html),
        Rule("/<project_name>/<version>/", endpoint=wsgi_html),
        Rule("/<project_name>/<version>/doap.rdf", endpoint=wsgi_rdf),
        Rule("/files/<project_name>/<version>/<download>", endpoint=wsgi_html),
    ])

    resource = WSGIResource(reactor, reactor, TWsgiApplication(url_map))
    site = Site(resource)

    reactor.listenTCP(PORT, site)

    logging.info('listening on: %s:%d' % (HOST, PORT))
    logging.info('wsdl is at: http://%s:%d/?wsdl' % (HOST, PORT))

    sys.exit(reactor.run())



# encoding: utf8
#
# (C) Copyright Arskom Ltd. <info@arskom.com.tr>
#               Uğurcan Ergün <ugurcanergn@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

import logging
logger = logging.getLogger(__name__)

import datetime
import os

from sqlalchemy import sql

from spyne.decorator import rpc
from spyne.error import ArgumentError
from spyne.error import ResourceNotFoundError
from spyne.model.primitive import String
from spyne.model.primitive import Unicode
from spyne.service import ServiceBase

from spyne.model.binary import File
from spyne.protocol.http import HttpPattern

from spynepi.const import FILES_PATH
from spynepi.db import Package
from spynepi.db import Release
from spynepi.db import Distribution
from spynepi.db import Person


class RootService(ServiceBase):
    @rpc(Unicode, Unicode, Unicode, Unicode, File, Unicode, Unicode, Unicode,
         Unicode, Unicode, Unicode, Unicode, Unicode, Unicode, Unicode,
         Unicode, Unicode, String, _patterns=[HttpPattern("/",verb="POST")])
    def register(ctx, name, license, author, home_page, content, comment,
            download_url, platform, description, metadata_version, author_email,
            md5_digest, filetype, pyversion, summary, version, protcol_version):

        exists = False
        pth = os.path.join("files", name, version)

        def generate_package():
            return Package(
                    package_name=name,
                    package_cdate=datetime.date.today(),
                    package_description=description,
                    rdf_about=os.path.join("/", name),
                    package_license=license,
                    package_home_page=home_page
                )

        def generate_person():
            return Person(person_name=author,
                    person_email=author_email,
                )

        def generate_release():
            return Release(
                    rdf_about=os.path.join("/", name, version),
                    release_version=version,
                    release_cdate=datetime.date.today(),
                    release_summary=summary ,
                    meta_version=metadata_version,
                    release_platform=platform,
                )

        def generate_dist():
            return Distribution(content_name=content.name,
                    content_path=pth,
                    dist_download_url=download_url,
                    dist_comment=comment,
                    dist_file_type=filetype,
                    dist_md5=md5_digest,
                    py_version=pyversion,
                    protocol_version=protcol_version,
                )

        def package_content():
            file = content
            path = os.path.join(FILES_PATH,pth)
            if os.path.exists(path):
                f = open(os.path.join(path,file.name),"w")
            else:
                os.makedirs(path)
                f = open(os.path.join(path,file.name),"w")

            for d in file.data:
                f.write(d)
            f.close()

        body = ctx.in_body_doc

        #TO-DO Add a method check
        check = ctx.udc.session.query(Package).filter_by(package_name=name).all()
        if check != []:
            exists = True
            for rel in check[0].releases:
                if rel.release_version == version and os.path.exists(pth) == True:
                    raise ArgumentError()

        if body[":action"][0] == "submit":
            if exists:
                check[0].releases.append(generate_release())

            else:
                package = generate_package()
                package.owners.append(generate_person())
                package.releases.append(generate_release())
                ctx.udc.session.add(package)
                ctx.udc.session.flush()
            ctx.udc.session.commit()

        if body[":action"][0] == "file_upload":
            if exists:
                rel = ctx.udc.session.query(Release).join(Package).filter(
                    sql.and_(
                        Package.package_name == name,
                        Release.release_version == version)
                    ).first()

                if rel is None:
                    raise ResourceNotFoundError(name)
                rel.distributions.append(generate_dist())
                package_content()

            else:
                package = generate_package()
                package.owners.append(generate_person())
                package.releases.append(generate_release())
                package.releases[-1].distributions.append(generate_dist())
                package_content()
                ctx.udc.session.add(package)

            ctx.udc.session.commit()




from collections import namedtuple

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

from spyne.model.complex import table
from spyne.model.complex import Array
from spyne.model.complex import TTableModel
from spyne.model.primitive import Date
from spyne.model.primitive import Integer32
from spyne.model.primitive import String
from spyne.model.primitive import Unicode

from spynepi.const import TABLE_PREFIX

DatabaseHandle = namedtuple("DatabaseHandle", ["db", "Session"])
TableModel = TTableModel()


class Person(TableModel):
    __tablename__ = "%s_person"  % TABLE_PREFIX
    __table_args__ = {"sqlite_autoincrement": True}

    id = Integer32(primary_key=True)
    person_name = String(60)
    person_email = String(60)


class Distribution(TableModel):
    __tablename__ = "%s_distribution"  % TABLE_PREFIX
    __table_args__ = {"sqlite_autoincrement": True}

    id = Integer32(primary_key=True)

    # TO-DO Add Content data
    content_name = String(256)
    content_path = String(256)
    dist_download_url = String(256)
    dist_comment = String(256)
    dist_file_type = String(256)
    dist_md5 = String(256)
    py_version = String(10)
    protocol_version = String(10)


class Release(TableModel):
    __tablename__ = "%s_release"  % TABLE_PREFIX
    __table_args__ = {"sqlite_autoincrement": True}

    id = Integer32(primary_key=True)
    release_cdate = Date
    rdf_about = String(256)
    release_version = String(10)
    meta_version = String(10)
    release_summary = String(256)
    release_platform = String(30)

    distributions = Array(Distribution).store_as(table(right="release_id"))


class Package(TableModel):
    __tablename__ = "%s_package"  % TABLE_PREFIX
    __table_args__ = {"sqlite_autoincrement": True}

    id = Integer32(primary_key=True)
    package_name = String(40)
    package_cdate = Date
    package_description = Unicode
    rdf_about = Unicode(256)
    package_license = Unicode(40)
    package_home_page = String(256)

    owners = Array(Person).store_as(table(right="owner_id"))
    releases = Array(Release).store_as(table(right="package_id"))


def init_database(connection_string):
    db = create_engine(connection_string)

    TableModel.Attributes.sqla_metadata.bind = db
    Session = sessionmaker(bind=db)

    # So that metadata gets updated with table names.
    import spynepi.entity.root
    import spynepi.entity.project

    TableModel.Attributes.sqla_metadata.create_all(checkfirst=True)

    return DatabaseHandle(db=db, Session=Session)



# encoding: utf8
#
# (C) Copyright Arskom Ltd. <info@arskom.com.tr>
#               Uğurcan Ergün <ugurcanergn@gmail.com>
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
# MA 02110-1301, USA.
#

import logging
logger = logging.getLogger(__name__)

import os
import shutil
import tempfile

from lxml import html
from sqlalchemy import sql

from spyne.error import ValidationError
from spyne.util import reconstruct_url

from pkg_resources import resource_filename

from sqlalchemy.orm.exc import NoResultFound

from spyne.decorator import rpc
from spyne.error import RequestNotAllowed
from spyne.model.binary import File
from spyne.model.complex import Array
from spyne.model.primitive import AnyUri
from spyne.model.primitive import Unicode
from spyne.protocol.html import HtmlPage
from spyne.protocol.http import HttpPattern
from spyne.service import ServiceBase

from spynepi.const import FILES_PATH
from spynepi.const import REPO_NAME
from spynepi.entity.project import Index
from spynepi.entity.project import Release
from spynepi.db import Package
from spynepi.db import Release


TPL_DOWNLOAD = os.path.abspath(resource_filename("spynepi.const.template",
                                                               "download.html"))


class IndexService(ServiceBase):
    @rpc (_returns=Array(Index), _patterns=[HttpPattern("/",verb="GET")])
    def index(ctx):
        return [Index(
                Updated=package.package_cdate,
                Package=AnyUri.Value(text=package.package_name,
                    href=package.releases[-1].rdf_about),
                Description=package.package_description,
            ) for package in ctx.udc.session.query(Package)]


def cache_package(spec, own_url):
    from glob import glob
    from setuptools.command.easy_install import main as easy_install
    from distutils.core import run_setup
    import ConfigParser

    path = tempfile.mkdtemp('.spynepi')
    easy_install(["--user", "-U", "--editable", "--build-directory",
                                                           path, spec])

    if os.environ.has_key('HOME'):
        rc = os.path.join(os.environ['HOME'], '.pypirc')
        config = ConfigParser.ConfigParser()

        if os.path.exists(rc):
            config.read(rc)

        try:
            config.add_section(REPO_NAME)

            config.set(REPO_NAME, 'repository', own_url)
            config.set(REPO_NAME, 'username', 'x')
            config.set(REPO_NAME, 'password', 'y')

        except ConfigParser.DuplicateSectionError:
            pass

        try:
            config.add_section('distutils')
        except ConfigParser.DuplicateSectionError:
            pass

        try:
            index_servers = config.get('distutils', 'index-servers')
            index_servers = index_servers.split('\n')
            if 'spynepi' not in index_servers:
                index_servers.append(REPO_NAME)

        except ConfigParser.NoOptionError:
            index_servers = [REPO_NAME]

        config.set('distutils', 'index-servers', '\n'.join(index_servers))

        config.write(open(rc,'w'))

    else: # FIXME: ??? No idea. Hopefully setuptools knows better.
        pass # raise NotImplementedError("$HOME not defined, .pypirc not found.")

    try:
        # plagiarized from setuptools
        setups = glob(os.path.join(path, '*', 'setup.py'))
        if not setups:
            raise ValidationError(
                "Couldn't find a setup script in %r editable distribution: %r" %
                                                    (spec, os.path.join(path,'*'))
            )

        if len(setups)>1:
            raise ValidationError(
                "Multiple setup scripts in found in %r editable distribution: %r" %
                                                    (spec, setups)
            )

        lib_dir =os.path.dirname(setups[0])
        os.chdir(lib_dir)
        dist = run_setup(setups[0])
        dist.commands = ['register', 'sdist', 'upload']
        dist.command_options = {
            'register': {'repository': ('command line', 'spynepi')},
            'upload': {'repository': ('command line', 'spynepi')},
            'sdist': {},
        }

        dist.run_commands()


    finally:
        shutil.rmtree(path)


class HtmlService(ServiceBase):
    @rpc(Unicode, Unicode, _returns=Unicode, _patterns=[
            HttpPattern("/<project_name>"),
            HttpPattern("/<project_name>/"),
            HttpPattern("/<project_name>/<version>"),
            HttpPattern("/<project_name>/<version>/"),
        ])
    def download_html(ctx, project_name, version):
        ctx.transport.mime_type = "text/html"

        try:
            ctx.udc.session.query(Package).filter_by(
                                                package_name=project_name).one()
        except NoResultFound:
            cache_package(project_name, reconstruct_url(ctx.transport.req_env,
                                                  path=False, query_string=False))

        download = HtmlPage(TPL_DOWNLOAD)
        download.title = project_name

        if version:
            release = ctx.udc.session.query(Release).join(Package).filter(
                sql.and_(
                    Package.package_name == project_name,
                    Release.release_version == version,
                    Package.id == Release.package_id
                )).one()

            download.link.attrib["href"] = "%s/doap.rdf" % (release.rdf_about)
            download.h1 = '%s-%s' % (project_name, version)

            download.a = release.distributions[0].content_name
            download.a.attrib["href"] = "/%s/%s#md5=%s" % (
                    release.distributions[0].content_path,
                    release.distributions[0].content_name,
                    release.distributions[0].dist_md5,
                )

        else:
            package = ctx.udc.session.query(Package) \
                                     .filter_by(package_name=project_name).one()

            download = HtmlPage(TPL_DOWNLOAD)
            download.link.attrib["href"] = '%s/doap.rdf' % (package.releases[-1].rdf_about)
            download.h1 = project_name
            download.a = package.releases[-1].distributions[0].content_name
            download.a.attrib["href"] = "/%s/%s#md5=%s" % (
                    package.releases[-1].distributions[0].content_path,
                    package.releases[-1].distributions[0].content_name,
                    package.releases[-1].distributions[0].dist_md5
                )

        return html.tostring(download.html)

    @rpc(Unicode, Unicode, Unicode, _returns=File, _patterns=[
                    HttpPattern("/files/<project_name>/<version>/<file_name>")])
    def download_file(ctx, project_name, version, file_name):
        repository_path = os.path.abspath(os.path.join(FILES_PATH,"files"))
        file_path = os.path.join(repository_path, project_name, version, file_name)
        file_path = os.path.abspath(file_path)

        if not file_path.startswith(repository_path):
            # This request tried to read data from where it's not supposed to
            raise RequestNotAllowed(repr([project_name, version, file_name]))

        return File.Value(name=file_name, path=file_path)
